// Copyright (c) 2002 Cathay Networks, Inc.  
// see LICENSE file for details of VOCAL license

#include <cassert>
#include <cstdio>
#include <cstring>
#include <errno.h>
#include <iostream.h>
#include <stdlib.h>

#ifdef WIN32

#include <winsock2.h>
#include <io.h>

#else

#include <arpa/inet.h>
#include <fcntl.h>
#include <netdb.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <unistd.h>

#endif

#include "network.h"

#ifndef WIN32 

int closesocket( Socket fd )
{
   if ( fd == INVALID_SOCKET)
   {
      return 0;
   }
    
   return close( fd );
}
#endif



/// Returns a file descript listening for inpcomin peer connections 
int 
createPeerListen(int port)
{
   //clog << "In Peer listen\n";
   int fd;
    
   fd = socket( AF_INET, SOCK_STREAM, 0);
   //clog << "socket fd is " << fd << endl;
   if ( fd == INVALID_SOCKET )
   {
      int e = errno;
      clog << "err1=" << e << endl;
      cerr << "Could not get a socket for the peer port:" 
           << strerror( e ) << endl;
      return INVALID_SOCKET;
   }
    
   // set reuse option 
   const int on=1;
   int optErr = setsockopt( fd, SOL_SOCKET, SO_REUSEADDR, (char*)&on, sizeof(on) );
   assert( optErr == 0 );
    
   struct sockaddr_in addr;
   memset( &addr,0, sizeof(addr) );
   addr.sin_family = AF_INET;
   addr.sin_addr.s_addr = htonl( INADDR_ANY );
   addr.sin_port = htons(port);
    
   int err = bind( fd,(sockaddr *)&addr, sizeof( addr) );
   if ( err != 0 )
   {
      int e = errno;
        
      switch (e)
      {
         case EADDRINUSE:
         {
            cerr << "Port " << port << " for listening for peers in in use" << endl;
            return INVALID_SOCKET;
         }
         break;
         default:
         {
            cerr << "Could not bind peer port:" << strerror( e ) << endl;
            return INVALID_SOCKET;
         }
         break;
      }
   }
    
   //clog << "Bind return " << err << endl;
   assert( err == 0);
    
   err = listen( fd, 3 );
   //clog << "listen return " << err << endl;
   assert( err == 0 );
    
   return fd;
}


/// takes the fd of listen port and returns a new fd of an accepted coneection
Socket
acceptPeer( Socket listenFd , int* ip, unsigned short* port )
{ 
   assert( ip );
   assert( port );
    
   //clog << "In acceptPeer " << endl;
    
   struct sockaddr_in peer;
   int peerLen = sizeof(peer);
    
   Socket fd = accept( listenFd, (struct sockaddr*) &peer, (socklen_t*) &peerLen );
   //clog << "accept socket fd is " << fd << endl;
   assert( fd != INVALID_SOCKET );
    
   *ip   = ntohl( peer.sin_addr.s_addr );
   *port = ntohs( peer.sin_port );

   // make non blocking 
#if WIN32
   // TODO - I think this code (with ioctlsocket changed to ioctl) will work under unix - test 
   unsigned long block = 0;
   int errNoBlock = ioctlsocket( fd, FIONBIO , &block );
   assert( errNoBlock == 0 );
#else
   int flags  = fcntl( fd, F_GETFL, 0);
   int errNoBlock = fcntl(fd,F_SETFL, flags| O_NONBLOCK );
   assert( errNoBlock == 0 );
#endif

   // set the no linger on close 
   struct linger opt;
   opt.l_onoff = 1;
   opt.l_linger = 0;
   int errOpt = setsockopt( fd, SOL_SOCKET, SO_LINGER, (char*)&opt, sizeof(opt));
   assert( errOpt == 0 );
    
   return fd;
}


/// parse host:port into an ip addres and port
void
parseHostName( char* peerName, int* ip, int* portVal )
{
   in_addr sin_addr;
    
   char host[512];
   strncpy(host,peerName,512);
   host[512-1]='\0';
   char* sep = strchr(host,':');
   char* port = NULL;
   char* defaultPort = "7000";
    
   if ( sep == NULL )
   {
      port = defaultPort;
   }
   else
   {
      *sep = '\0';
      port = sep+1;
   }
    
   // figure out the host part 
   struct hostent* h;
   h = gethostbyname( host );
   if ( h == NULL )
   {
      cerr << "Unknown host: " << host << endl;
      exit(1);
   }
    
   sin_addr = *(struct in_addr*)h->h_addr;
   *ip = ntohl( sin_addr.s_addr );
    
   //char* ipString = inet_ntoa( sin_addr );
   //clog << "    resolved to IP of " << ipString << endl;
    
   // set the port part 
   char* endPtr=NULL;
   int portNum = strtol(port,&endPtr,0);
   //clog << "Got a port of " << portNum << endl;
   if ( endPtr != NULL )
   {
      if ( *endPtr != '\0' )
      {
         cerr << " could not parse port number" << endl;
         exit(1);
      }
   }
    
   assert( portNum >= 1024 );
   assert( portNum <= 65000 );
   *portVal = portNum;
}


/// connect to a peer - returns file descriptor of connection - takes name of peer
Socket
connectPeer( int ip, unsigned short portNum, unsigned short srcPortBase )
{
   Socket fd;
   int srcPort=0;
    
   struct sockaddr_in addr;
   memset(&addr,0,sizeof(addr));
   addr.sin_family = AF_INET;
    
   addr.sin_addr.s_addr = htonl(ip);
   addr.sin_port = htons( portNum );
    
   fd = socket( AF_INET, SOCK_STREAM, 0 );
   assert( fd != INVALID_SOCKET );

   // find a src port to use 
   if ( srcPortBase != 0 )
   {
      int port = srcPortBase; 

      while (srcPort==0)
      {
         //cerr << "Try port " << port << endl;
	    
         struct sockaddr_in srcAddr;
         memset(&srcAddr,0,sizeof(srcAddr));
         srcAddr.sin_family = AF_INET;
         srcAddr.sin_addr.s_addr = htonl(INADDR_ANY);
         srcAddr.sin_port = htons( port );
         int errBind =  bind(fd,(struct sockaddr*)&srcAddr, 
                             (socklen_t)(sizeof(srcAddr)));
         if ( errBind != 0 )
         {
            //int e = errno;
            //cerr << "Bind or peer source port got error: " 
            //     << strerror(e) 
            //     << " on src port " << srcPort <<endl;
         }
         else
         {
            // got one that worked - use it 
            srcPort = port;
         }

         port++;
         if ( port >= srcPortBase + 500 )
         {
            cerr << "Could not find a free port to use "
                 << "for source of peer connection" << endl;
            srcPort=-1;
         }
      }
   }
    
   clog << "Trying connect to " << ip << ":"<< portNum<< endl;
   int err = connect( fd, (struct sockaddr*)&addr, sizeof(addr) );
   if ( err != 0 )
   {
      int e = errno;
        
      switch( e )
      {
         case ECONNREFUSED: 
            cerr << "Connection to peer refused" << endl;
            break;
         case EHOSTUNREACH:
            cerr << "Peer host is unreachable" << endl;
            break;
         default:
            clog << "Connect returned err " << e << " " << strerror(e) << endl;
            cerr << "Could not connect to peer " << ip << ":"<< portNum  << endl;
      }
      return INVALID_SOCKET;
   }
   clog << "Got connection" << endl;
    
   // make non blocking 
#if WIN32
   unsigned long block = 0;
   int errNoBlock = ioctlsocket( fd, FIONBIO , &block );
   assert( errNoBlock == 0 );
#else
   int flags  = fcntl( fd, F_GETFL, 0);
   int errNoBlock = fcntl(fd,F_SETFL, flags| O_NONBLOCK );
   assert( errNoBlock == 0 );
#endif

   struct linger opt;
   opt.l_onoff = 1;
   opt.l_linger = 0;
   int errOpt = setsockopt( fd, SOL_SOCKET, SO_LINGER, (char*)&opt, sizeof(opt));
   assert( errOpt == 0 );
    
   cerr << "Connect to peer " 
        <<  ip << ":"<< portNum 
        << " from port " << srcPort << endl;
   return fd;
}

// Local Variables:
// mode:c++
// c-file-style:"bsd"
// c-file-offsets:((case-label . +))
// c-basic-offset:4
// indent-tabs-mode:nil
// End:
