// Copyright (c) 2002 Cathay Networks, Inc.  
// see LICENSE file for details of VOCAL license

/**************** Manage Peer connection *************************/
#include <cassert>
#include <cstdio>
#include <cstring>
#include <errno.h>
#include <iostream.h>

#ifdef WIN32

#include <winsock2.h>
#include <stdlib.h>
#include <io.h>

#else

#include <getopt.h>
#include <netinet/in.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <unistd.h>

#endif




#include "config.h"
#include "peer.h"
#include "probe.h"
#include "rib.h"
#include "stats.h"
#include "time.h"

Peer peer[maxNumPeers];
int numPeers=0;

/// Shutdown the connetion to a peer 
void
closePeer( int p )
{
   if ( peer[p].fd == INVALID_SOCKET )
   {
      return;
   }

   stats.numPeerCloses++;
    
   cerr << "Closing conneciton with peer " << peer[p].peerName << endl;
    
   int fd = peer[p].fd;
    
   peer[p].fd = INVALID_SOCKET;
   peer[p].synced = false;
    
   closesocket( fd );
}


/// send an update of a single element in the rib to a particular peer
void
sendUpdate( int ribIndex, int peerIndex )
{
   Socket fd = peer[peerIndex].fd;
    
   if ( fd == INVALID_SOCKET ) 
   {
      // peer is not up
      return;
   }
    
   int proxy = getRibProxy( ribIndex );
   assert( proxy >= 0 );
   assert( proxy < numDstProxies );
    
   int proxyIp = dstProxy[proxy].ip;
   int proxyPort = dstProxy[proxy].port;
   Int64 time = rib[ribIndex].time;
    
   assert( peerIndex >= 0 );
   assert( peerIndex < numPeers );
    
   int hashId = rib[ribIndex].hash; 
    
   Update update;
   update.type = 1;
   update.size = sizeof(update);
   update.hash = htonl(hashId);
   update.port = htons(proxyPort);
   update.ip   = htonl(proxyIp);

   assert( time >= 0 );
   Int32 timeLow = low32(time);
   Int32 timeHigh = high32(time);
   update.time.low  = htonl( timeLow );
   update.time.high = htonl( timeHigh );
    
   //cerr << "Sent hash of " << hex << hashId << endl;
   //cerr << "Sent time of " << hex << time << endl;
    
   int l = send(fd,(char*)&update,sizeof(update),0);
   if ( l != sizeof(update ) )
   {
      int e = errno;
        
      if ( l == -1 )
      {
         switch (e)
         {
            case ECONNRESET:
               cerr << "Connection to peer got reset" << endl;
               break;
            default:
               cerr << "Send update to peer had error " << e << endl;
               break;
         }
      }
      else
      {
         cerr << "Send update to peer returned " << l << endl;
      }
        
      // problem with connection to peer
      closePeer( peerIndex );
      return;
   }

   stats.updatesSent++;
}


/// get an update from a peer and flood it to other peers 
void 
receiveUpdate( Socket fd )
{
   Update update;
    
   int peerIndex = findPeer( fd );
   assert( peerIndex != -1 );
    
   // receive it 
   char* buf = reinterpret_cast<char*>(&update);
   int num = sizeof(update);

   while ( num > 0 )
   {
      int l = recv(fd,buf,num,0);
      if ( l == -1 )
      {
         int e = errno;

         switch (e)
         {
            case ECONNRESET:
               // sceilently ignore 
               break;
            default:
            {
               cerr << "Problem receiving peer update data "
                    << " returned " << l << " error=" << e << endl;
            }
            break;
         }
	    
         // problem with connection to peer
         closePeer( peerIndex );
         return;
      }   
      buf += l;
      num -= l;
   }

   if ( update.type != 1 )
   {
      cerr << "Received bad message from peer" << endl;
      closePeer( peerIndex );
      return;
   }
    
   assert( update.size == sizeof(update) );
   int hash = ntohl(update.hash);
   int proxyPort = ntohs(update.port); 
   int proxyIp = ntohl( update.ip );
   Int32 timeLow  = ntohl(update.time.low );
   Int32 timeHigh = ntohl(update.time.high);
   Int64 time = form64(timeLow,timeHigh);

   //cerr << "Got hash of " << hex << hash << endl;
   //cerr << "Got time of " << hex << time << endl;
    
   assert( time >= 0 ); // TODO change to error instead of assert
    
   int proxy = findProxy( proxyIp, proxyPort );
   if ( proxy == -1 )
   {
      // learned about a new proxy - need to add it. 
      addProxy( proxyIp, proxyPort,0,0,true );
        
      proxy = findProxy( proxyIp, proxyPort );
      assert( proxy != -1 );
   }
    
   // see if it updates the rib
   int ribIndex = getRibIndex( hash );
   if (ribIndex!=-1)
   {
      if ( rib[ribIndex].proxy != proxy )
      {
         //cerr << "Conflict in hash table on sync\n";
         stats.numSyncConflicts++;
            
         if ( time > rib[ribIndex].time )
         {
            // update wins 
            rib[ribIndex].proxy=-1;
            rib[ribIndex].time=0;
         }
         else
         {
            // local wins 
            time = rib[ribIndex].proxy;
            proxy = rib[ribIndex].proxy;
            fd = INVALID_SOCKET; // forget person who sent it so it gets flooded to everyone 
                
            // flood update to peers other than the one we recieved it from
            for (int p=0; p<numPeers; p++ )
            {
               if ( peer[p].fd != fd  )
               {
                  sendUpdate( ribIndex, p );
               }
            }
                
            return;
         }
      }
   }
    
   if ( time <= rib[ribIndex].time )
   {
      // this update is older than current - ignore 
      return;
   }
    
   // do the update 
   rib[ribIndex].proxy = proxy;
   rib[ribIndex].time = time;
    
   for (int p=0; p<numPeers; p++ )
   {
      if ( ( peer[p].fd != INVALID_SOCKET ) && ( peer[p].fd != fd ) )
      {
         int l = write(peer[p].fd,&update,sizeof(update));
         if ( l != sizeof(update ) )
         {
            closePeer(p);
         }
      }
   }

   stats.updatesReceived++;
}


/// tries to connect to all the peers that we currently don't have a connection too
void 
processPeers()
{
   int p;
    
   // try to connect to all peers 
   for (p=0; p<numPeers; p++)
   {
      if (  (peer[p].fd==INVALID_SOCKET) && (peer[p].peerName[0]!='\0') )
      {
         // only try to conect if aproperate time has passed since last try
         Int64 now = getTime();
            
         if ( peer[p].lastConnectTime + 10*oneSecond < now )
         {
            peer[p].lastConnectTime = now;
            int portNum;
            int ip;

            parseHostName( peer[p].peerName, &ip, &portNum );

            peer[p].fd = connectPeer( ip, portNum, setting.peerListenPort );

            peer[p].id = form64(portNum,ip);
            peer[p].syncedRecords = 0;
            peer[p].synced = false;
         }
      }
   }
    
   // if not fully synced with peer, send a block of sync data  
   int work=0; // count of records sent in this call - used to stop after some time
   int maxWork = 4096/sizeof(Update);// TODO make this a config option number 

   for (p=0; p<numPeers; p++)
   {
      // if we have a conection but not synced yet 
      if ( (peer[p].fd!=INVALID_SOCKET) && (peer[p].synced == false ) ) 
      {
         int i = peer[p].syncedRecords;
         //clog << "start sync at record " << i << endl;
         while ( i<ribSize )
         {
            expireRib(i);
            if ( rib[i].proxy != -1 )
            {
               // send this itme to peer
               sendUpdate( i, p );
               work++;
                    
               if ( work > maxWork )  
               {
                  // run out of time in this time slice, punt 
                  peer[p].syncedRecords = i;
                  //clog << "Progress syncing to " << i << " with " << peer[p].peerName << endl;
                  return;
               }
                    
            }                
            i++;   
         }
         //clog << "completed sync at record " << i << " work was " << work << endl;
         peer[p].syncedRecords = i;
         peer[p].synced = true;
            
         if ( peer[p].peerName[0] == '\0' )
         {
            cerr << "Completed sync with peer number " << p << endl;
         }
         else
         {
            cerr << "Completed sync with " << peer[p].peerName << endl;
         }
      }
   }
}


/// a new peer connected to us, manage it 
void
addPeer( Socket peerListenFd )
{
   stats.peersAdded++;
    
   int ip;
   unsigned short port;
    
   Socket peerFd = acceptPeer( peerListenFd, &ip, &port );
   assert( peerFd != INVALID_SOCKET );

   cerr << "New peer connection from " << ip << ":" << port << endl;
    
   // round port to lowe 1000
   port = port/1000;
   port = port*1000;
    
   // TODO - not sure this port is right way to match incoming peer conections 
   Int64 id = form64(port,ip);
    
   // is this a peer we know about 
   int index=-1;
   for ( int i=0; i<numPeers; i++ )
   {
      if ( peer[i].id == id )
      {
         cerr << "new peer is replacing peer number " << i << endl;
         index = i;
         break;
      }
   }
    
   // this is a newly discovered peer - add to end of peer list 
   if ( index == -1 )
   {
      if ( numPeers+1 >= maxNumPeers )
      {
         cerr << "Max number of peers exceeded - ignoring new connection from " 
              << ip << ":" << port << endl;
         return;
      }
      index = numPeers++;
      peer[index].fd = INVALID_SOCKET;
   }
    
   Peer& p = peer[index];
    
   // close the old connection if it exists 
   if ( p.fd != INVALID_SOCKET ) 
   {
      closesocket( p.fd ); p.fd = INVALID_SOCKET;
   }
    
   p.fd = peerFd;
   p.lastConnectTime = getTime();
   p.synced = false;
   p.syncedRecords = 0;
   p.peerName[0] = '\0';
   p.id = id;

   cerr << "Done adding new peer" << endl;
}

// Local Variables:
// mode:c++
// c-file-style:"bsd"
// c-file-offsets:((case-label . +))
// c-basic-offset:4
// indent-tabs-mode:nil
// End:
