// Copyright (c) 2002 Cathay Networks, Inc.  
// see LICENSE file for details of VOCAL license

/***************************** SIP message handling *********************/
#include <cassert>
#include <iostream.h>
#include <string.h>
#include <cstdio>
#include <stdlib.h>

#include "config.h"
#include "peer.h"
#include "probe.h"
#include "rib.h"
#include "sip.h"
#include "stats.h"
#include "string.h"
#include "time.h"
#include "udp.h"


/// deal with a SIP response to a transaction 
void
handleResponse( Socket fd, Socket sendFd )
{
   stats.numResponses++;
    
   //clog << "In handleResponse" << endl;
    
   char buf[udpMaxMessageLength];
   int len;
   unsigned int srcIp;
   unsigned short srcPort;
    
   if ( getMessage(fd,buf,&len,&srcIp,&srcPort) )
   {
      //clog << "Response: " << len << " bytes of " << buf <<endl;
        
      // find the top via 
      int viaStart = findHeader("Via",buf,len);
      if ( viaStart >= len )
      {
         cerr << "Can't find Via header in incoming response" << endl;
         return;
      }
        
      // find the via tag
      int tagStart = viaStart + findBranch(buf+viaStart,len-viaStart);
      if ( tagStart >= len )
      {
         cerr << "Can't find tag in incoming response" << endl;
         return;
      }
      int tagEnd = viaStart + findEOL(buf+viaStart,len-viaStart);
      if ( tagEnd >= len )
      {
         cerr << "Can't find the end of the via header to pop" << endl;
         return;
      }
        
      // find where this message came from
      int ip,port;
      sscanf(buf+tagStart+strlen("branch="),"%x.%x",&ip,&port);
      //clog << "need to send response to ip " << hex << ip << dec << ":" <<port << endl;
        
      // form the new message with popped via 
      char* msgVec[2];
      int msgLen[2];
      msgVec[0] = buf;
      msgVec[1] = buf+tagEnd+2;
      msgLen[0] = viaStart;
      msgLen[1] = len-tagEnd-2;
        
      // send this message back to where the response came from
      sendMessage( sendFd, msgVec, msgLen, 2, ip, port );
   }
}


/// select one of the proxies that is up to use for a new call
int 
selectProxy(unsigned int random)
{
   int priority = -1;
   int i;
    
   // find max priority of any proxies that are up 
   for ( i=0; i<numDstProxies; i++ )
   {
      if ( dstProxy[i].up )
      {
         priority = max( priority, dstProxy[i].priority );
      }
   }
    
   if ( priority == -1 )
   {
      // no proxies are up 
      return -1;
   }
    
   // find sum of weights of all up proxies that are up and have this priority 
   unsigned int wSum=0;
   int num=0;
   int proxy=-1;
    
   for ( i=0; i<numDstProxies; i++ )
   {
      if ( dstProxy[i].up  && ( dstProxy[i].priority == priority ) )
      {
         wSum += dstProxy[i].weight;
         proxy = i;
         num++;
      }
   }
    
   if ( num == 1 )
   {
      // only one possibility - use it 
      return proxy;
   }
    
   if (random == 0 )
   {
      // TODO get a random number somehow
   }
   random = random%wSum;
    
   wSum=0;
   for ( i=0; i<numDstProxies; i++ )
   {
      if ( dstProxy[i].up  && ( dstProxy[i].priority == priority ) )
      {
         wSum += dstProxy[i].weight;
         if ( wSum > random )
         {
            return i;
         }
      }
   }
    
   assert(0);
   return -1;
}


/// deal with a new SIP request that starts a transaction 
void
handleRequest( Socket fd )
{
   stats.numRequests++;
    
   //clog << "In handleRequest" << endl;
    
   char buf[udpMaxMessageLength];
   int len;
   unsigned int srcIp;
   unsigned short srcPort;
    
   if ( getMessage(fd,buf,&len,&srcIp,&srcPort) )    
   {
      //clog << "Request: " << len << " bytes of " << buf 
      // << "from 0x" << hex << srcIp << dec << " port " << srcPort <<endl;
      unsigned int hash=0;
        
     // figure out the data to hash in the sticky header  
      int callIdStart = findHeader(setting.useStickyType,buf,len);
      int callIdEnd = callIdStart + findEOL(buf+callIdStart,len-callIdStart);
        
#if 1 
       clog << "Sticky header data is: ";
      for ( int i=callIdStart; i<callIdEnd; i++ ) clog << buf[i];
      clog << endl;
#endif
        
      hash = stringHash(buf+callIdStart,callIdEnd-callIdStart);
        
      // figure out which proxy to send this to
      int proxy=-1;
      // look for it in the RIB
      proxy = getRibProxy( hash );
      if ( proxy == -1 )
      {
         proxy = selectProxy(hash);
      }
      if ( proxy == -1 )
      {
         // no proxies availabel to send message 
         cerr << "No proxies are up - can not send message to anyone" << endl;
         return;
      }
        
      // update the sticky cache to use this proxy 
      int ribIndex = getRibIndex(hash);
      rib[ribIndex].proxy = proxy;
      rib[ribIndex].time = getTime();
      assert( rib[ribIndex].hash == hash );
      for (int peer=0; peer<numPeers; peer++ )
      {
         sendUpdate(ribIndex,peer);
      }
        
     // find the top most via
      int viaStart = findHeader("Via",buf,len);

      // figure out the port in the via 
      int viaPort = 5060; // currently hardcoded to port 5060
      if ( setting.supportNat )
      {
          // when it came through a NAT, should reply to port it came from and ignore what via says 
          viaPort = srcPort;
      }
      else
      {
          // parse the via port out of the via header
          const int viaLen = 4;
          int colon = findColon(buf+viaStart+viaLen,len-viaLen-viaStart)+viaStart+viaLen;
          int eol = findEOL(buf+viaStart,len-viaStart)+viaStart;
    
           if ( (colon!=viaStart+viaLen) && ( colon < eol ) )
           {
               int p = atoi(buf+colon+1);
               if ( p == 0 )
               {
                   // some error in parsing port
                   cerr << "Problem in parsing via port on incoming request"
                        << "dropping the request" << endl;

                   return;
               }
               viaPort = p; 
               //clog << "Parsed out a via port of " << viaPort << endl;
           }
           else
           {
               // there is no port in the via line - use the sip defalt
               viaPort = 5060;
           }
      }

      // figure the new via line 
      char newVia[256];
      
      // TODO - prebuild this via 
      sprintf(newVia,"Via: SIP/2.0/UDP %s:%d;branch=%x.%hx\r\n",
              setting.proxyName,setting.respPort,srcIp,viaPort);
      int newViaLen=strlen(newVia);
      //clog << "New via is " << newVia << endl;
               
      // insert the new via and send message to destination proxy
      char* msgVec[3];
      int msgLen[3];
      msgVec[0] = buf;
      msgVec[1] = newVia;
      msgVec[2] = buf+viaStart;
      msgLen[0] = viaStart;
      msgLen[1] = newViaLen;
      msgLen[2] = len - viaStart;
        
      if ( dstProxy[proxy].fd == INVALID_SOCKET )
      {
         dstProxy[proxy].fd = openSendPort( dstProxy[proxy].ip, dstProxy[proxy].port);
      }
        
      if ( dstProxy[proxy].fd == INVALID_SOCKET )
      {
         sendMessage( dstProxy[proxy].fd, msgVec, msgLen, 3, 
                      dstProxy[proxy].ip, dstProxy[proxy].port );
      }
      else
      {
         sendMessage( dstProxy[proxy].fd, msgVec, msgLen, 3);
      }
   }
}

// c-basic-offset:4
// Local Variables:
// mode:c++
// c-file-style:"ellemtel"
// c-file-offsets:((case-label . +))
// indent-tabs-mode:nil
// End:
