// Copyright (c) 2002 Cathay Networks, Inc.  
// see LICENSE file for details of VOCAL license

#include <cassert>
#include <time.h>

#ifdef WIN32
#  include <errno.h>
#  include <winsock2.h>
#  include <stdlib.h>
#else
#  include <sys/time.h>
#endif

#include "time.h"

/******************* TIME functions *******************************/

/// returns the current time in ms since NTP epoc as a 64 bit integer
Int64
getTimeSlow()
{
   Int64 ret;
#ifdef WIN32 
   SYSTEMTIME epocSysTime;
    
   memset( &epocSysTime, 0 , sizeof(epocSysTime) );
   epocSysTime.wYear = 1900;
   epocSysTime.wMonth = 1;
   epocSysTime.wDay = 1;
   FILETIME epocFileTime;

   BOOL err3 = SystemTimeToFileTime(&epocSysTime, &epocFileTime);  
   assert( err3 != 0 );
   LARGE_INTEGER epocLarge;
   epocLarge.LowPart = epocFileTime.dwLowDateTime;
   epocLarge.HighPart = epocFileTime.dwHighDateTime;
   Int64 epoc = epocLarge.QuadPart;

   SYSTEMTIME sysTime;
   GetSystemTime(&sysTime);

   FILETIME time; // This is UTC in 100ns steps since 1601 
   GetSystemTimeAsFileTime( &time ); 

   ULARGE_INTEGER lint;
   lint.LowPart = time.dwLowDateTime;
   lint.HighPart = time.dwHighDateTime;

   ret = (lint.QuadPart-epoc)/10; // time is ms since epoc 
#else
   const Int64 UNIX_TO_NTP_OFFSET = (Int64)(70 * 365 + 17) * (Int64)(86400);

   time_t t;
   // unix time_t is since 1970 
   time(&t);
    
   ret = t;
   ret += UNIX_TO_NTP_OFFSET; // move epoc from 1970 to 1900 
   ret *= oneSecond; // conver from seconds to ms 
#endif
   return ret;
}
    

///
Int64
getTimeRelative()
{
   Int64 ret;
#ifdef WIN32
   static bool initialized=false;
   static Int64 scale=1;
    
   if ( !initialized )
   {
      LARGE_INTEGER freq;
      BOOL err2 = QueryPerformanceFrequency( &freq );
      assert( err2 != 0 );
      initialized = true;
        
      scale = freq.QuadPart/oneSecond;
   }
    
   LARGE_INTEGER perfTime;
   BOOL err1 = QueryPerformanceCounter( &perfTime );
   assert( err1 != 0 );
    
   ret = perfTime.QuadPart/scale;
#else
   struct timeval tv;
   int e = gettimeofday(&tv,NULL);
   assert( e == 0 );
    
   assert( oneSecond == 1000000 );
    
   ret = tv.tv_sec;
   ret *= 1000000;
   ret += tv.tv_usec;
#endif

   return ret;
}


/// returns the current time in micro seconds since NTP epoc as a 64 bit integer
Int64
getTime(bool update)
{
   static Int64 lastTime=0;
   if ( !update )
   {
      return lastTime;
   }
    
   static bool initialized=false;
   static Int64 offset;

   if (!initialized)
   {
      Int64 old=getTimeSlow();    
      Int64 slow=old;
      while( old == slow )
      {  
         slow = getTimeSlow();
      }
      Int64 rel = getTimeRelative();
      offset = slow-rel;
      initialized=true;
      //clog << "time offset=" << (int)(offset>>32) << " " << (int)(offset) << endl;
   }

   Int64 ret = getTimeRelative();
    
   ret += offset;

   assert( ret > 0 );
    
   lastTime = ret;
    
   return ret;
}


/// get a NTP reprsentation of time 
Int64
getNtpTime()
{
   // NTP is 64 bit fixed point 
   // first 32 bit is seconds since 1900 UT
   // second 32 bits is fractional part of seconds 

   Int64 t = getTime();
   assert( oneSecond == 1000000 );
   unsigned int seconds = low32(t/oneSecond);
   unsigned int usec    = low32(t%oneSecond);

   // based on idea that 2^32/10^6 = 4096 + 256 - 1825/32 
   unsigned int temp = (usec * 1825) >> 5;
   unsigned int low = ((usec << 12) + (usec << 8) - temp);

   Int64 ret = ( (Int64)(seconds)<<32 ) | low; 

   return ret;
}

// Local Variables:
// mode:c++
// c-file-style:"bsd"
// c-file-offsets:((case-label . +))
// c-basic-offset:4
// indent-tabs-mode:nil
// End:
