// Copyright (c) 2002 Cathay Networks, Inc.  
// see LICENSE file for details of VOCAL license

/*************************** UDP networking stuff  *******************/
#include <cassert>
#include <cstdio>
#include <cstring>
#include <errno.h>
#include <iostream.h>

#ifdef WIN32

#include <winsock2.h>
#include <stdlib.h>
#include <io.h>

#else

#include <stdlib.h>
#include <unistd.h>
#include <getopt.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/types.h>

#include <string.h>

#endif


#include "string.h"
#include "network.h"
#include "udp.h"


/// Open a UDP socket to send to a particular destination
Socket
openSendPort( int ip, unsigned short port )
{
   Socket fd;
    
   fd = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
   if ( fd < 0 )
   {
      assert(0);
   }
    
   struct sockaddr_in addr;
    
   memset((char*) &(addr),0, sizeof((addr)));
   addr.sin_family = AF_INET;
   addr.sin_addr.s_addr = htonl(ip);
   addr.sin_port = htons(port);
    
   if ( connect( fd,(struct sockaddr*)&addr, sizeof(addr)) != 0 )
   {
      int err = errno;
	
      cerr << "Got error: " << strerror(err) 
           << " while trying to connect a UDP socket to port " 
           << port << " on host " << ip << endl;

      return INVALID_SOCKET;
   }
    
   // make non blocking 
#if WIN32
   unsigned long block = 0;
   int errNoBlock = ioctlsocket( fd, FIONBIO , &block );
   assert( errNoBlock == 0 );
#else
   int flags  = fcntl( fd, F_GETFL, 0);
   int errNoBlock = fcntl(fd,F_SETFL, flags| O_NONBLOCK );
   assert( errNoBlock == 0 );
#endif

   //clog << "Opened UDP socket to " << ip <<":"<< port << " on fd " << fd << endl;
   return fd;
}


/// Open a UDP socket to receive on the given port 
Socket
openRecvPort( unsigned short port )
{
   if ( port == 0 )
   {
      return INVALID_SOCKET;
   }
    
   Socket fd;
    
   fd = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
   if ( fd == INVALID_SOCKET )
   {
      int err = errno;
      cerr << "Could not create a UDP socket:" << err << endl;
      return INVALID_SOCKET;
   }
    
   struct sockaddr_in addr;
   memset((char*) &(addr),0, sizeof((addr)));
   addr.sin_family = AF_INET;
   addr.sin_addr.s_addr = htonl(INADDR_ANY);
   addr.sin_port = htons(port);
    
   if ( bind( fd,(struct sockaddr*)&addr, sizeof(addr)) != 0 )
   {
      int e = errno;
        
      switch (e)
      {
         case EADDRINUSE:
         {
            cerr << "Port " << port << " for receiving UDP is in use" << endl;
            return INVALID_SOCKET;
         }
         break;
         default:
         {
            cerr << "Could not bind UDP receive port:" << strerror( e ) << endl;
            return INVALID_SOCKET;
         }
         break;
      }
   }
    
   // make non blocking 
#if WIN32
   unsigned long block = 0;
   int errNoBlock = ioctlsocket( fd, FIONBIO , &block );
   assert( errNoBlock == 0 );
#else
   int flags  = fcntl( fd, F_GETFL, 0);
   int errNoBlock = fcntl(fd,F_SETFL, flags| O_NONBLOCK );
   assert( errNoBlock == 0 );
#endif

   //clog << "Opened port " << port << " on fd " << fd << endl;
    
   return fd;
}


// Via: SIP/2.0/UDP 192.168.2.51:5060
/// recive a UDP message 
bool 
getMessage( Socket fd, char* buf, int* len,
	    unsigned int* srcIp, unsigned short* srcPort )
{
   struct sockaddr_in from;
   int fromLen = sizeof(from);
    
   *len=recvfrom(fd,buf,udpMaxMessageLength,0,
                 (struct sockaddr *)&from,(socklen_t *)&fromLen);
    
   if ( *len == -1 )
   {
      return false;
   }
    
   if ( *len == 0 )
   {
      return false;
   }
    
   *srcPort = ntohs(from.sin_port);
   *srcIp = ntohl(from.sin_addr.s_addr);
   // should append this information into the received= tag of the via
   
   if ( (*len)+1 >= udpMaxMessageLength )
   {
      cerr << "Received a message that was too large" << endl;
      return false;
   }
   buf[*len]=0;
    
   return true;
}


/// send a UDP message 
bool 
sendMessage( Socket fd, char* msgVec[], int lenVec[], int numVec, 
	     unsigned int dstIp, unsigned short dstPort )
{
   // TODO - make this work for connected UDP sockets
    
   assert( fd != INVALID_SOCKET );
   assert( numVec >= 1 );
    
   char buf[udpMaxMessageLength];
   int l=0;
    
   for( int i=0; i<numVec; i++ )
   {
      if ( l + lenVec[i] >= udpMaxMessageLength-1 )
      {
         cerr << "Resulting message is too large" << endl;
         return false;
      }
        
      //cerr << "msg " << i << " is len=" << lenVec[i] << " ->" << msgVec[i] << "<-" << endl;
        
      memcpy(buf+l,msgVec[i],lenVec[i]);
      l += lenVec[i];
   }
    
   assert( l < udpMaxMessageLength );
   buf[l]=0;
    
   int s;
   
   if ( dstPort == 0 )
   {
      // sending on a connected port 
      assert( dstIp == 0 );
        
      //clog << "send msg to connected" << endl;

      s = send(fd,buf,l,0);
   }
   else
   {
      assert( dstIp != 0 );
      assert( dstPort != 0 );
        
      struct sockaddr_in to;
      int toLen = sizeof(to);
      memset(&to,0,toLen);
        
      to.sin_port = htons(dstPort);
      to.sin_addr.s_addr = htonl(dstIp);
        
#if 0
      unsigned int d3 = (dstIp>>24) & 0xFF;
      unsigned int d2 = (dstIp>>16) & 0xFF;
      unsigned int d1 = (dstIp>> 8) & 0xFF;
      unsigned int d0 = (dstIp    ) & 0xFF;
      clog << "sendmsg to "<< d3<<"."<<d2<<"."<<d1<<"."<<d0<<":"<<dstPort<< endl;
      for ( int j=0; j<l; j++ ) 
      {
         clog << buf[j];
      }
      clog << endl;
#endif
        
      s = sendto(fd, buf, l, 0,(sockaddr*)&to, toLen);
   }
    
   if ( s == SOCKET_ERROR )
   {
      int e = errno;
      switch (e)
      {
         case ECONNREFUSED:
         case EHOSTDOWN:
         case EHOSTUNREACH:
         {
            // scielently ignore this 
         }
         break;
         default:
         {
            cerr << "err " << e << " "  << strerror(e) << " in send" << endl;
         }
      }
      return false;
   }
    
   if ( s == 0 )
   {
      cerr << "no data sent in send" << endl;
      return false;
   }
    
   if ( s != l )
   {
      cerr << "only " << s << " out of " << l << " bytes sent" << endl;
      return false;
   }
    
   return true;
}

// Local Variables:
// mode:c++
// c-file-style:"ellemtel"
// c-file-offsets:((case-label . +))
// indent-tabs-mode:nil
// End:
